from typing import Any

from pydantic import Field

from handbrake.models.common import Duration, Fraction, HandBrakeModel


class AudioAttributes(HandBrakeModel):
    alt_commentary: bool
    commentary: bool
    default: bool
    normal: bool
    secondary: bool
    visually_impaired: bool


class Audio(HandBrakeModel):
    attributes: AudioAttributes
    bit_rate: int
    channel_count: int
    channel_layout: int
    channel_layout_name: str
    codec: int
    codec_name: str
    codec_param: int
    description: str
    lfe_count: int = Field(alias="LFECount")
    language: str
    language_code: str
    sample_rate: int
    track_number: int


class Chapter(HandBrakeModel):
    duration: Duration
    name: str


class Color(HandBrakeModel):
    bit_depth: int = -1
    chroma_location: int
    chroma_subsampling: str = ""
    format: int
    matrix: int
    primary: int
    range: int
    transfer: int


class Geometry(HandBrakeModel):
    height: int
    par: Fraction = Field(alias="PAR")
    width: int


class SubtitleAttributes(HandBrakeModel):
    four_by_three: bool = Field(alias="4By3")
    children: bool
    closed_caption: bool
    commentary: bool
    default: bool
    forced: bool
    large: bool
    letterbox: bool
    normal: bool
    pan_scan: bool
    wide: bool


class Subtitle(HandBrakeModel):
    attributes: SubtitleAttributes
    format: str
    language: str
    language_code: str
    source: int
    source_name: str
    track_number: int


class Title(HandBrakeModel):
    angle_count: int
    audio_list: list[Audio]
    chapter_list: list[Chapter]
    color: Color
    crop: tuple[int, int, int, int]
    duration: Duration
    frame_rate: Fraction
    geometry: Geometry
    index: int
    interlace_detected: bool
    loose_crop: tuple[int, int, int, int] = (-1, -1, -1, -1)
    metadata: dict[str, Any]
    name: str
    path: str
    playlist: int
    subtitle_list: list[Subtitle]
    type: int
    video_codec: str


class TitleSet(HandBrakeModel):
    main_feature: int
    title_list: list[Title]



class HandBrakeError(Exception):
    def __init__(self, return_code: int):
        super().__init__()
        self.return_code = return_code

    def __str__(self) -> str:
        return "handbrake exited with return code " + str(self.return_code)



import subprocess
from contextlib import ExitStack
from os import PathLike
from tempfile import NamedTemporaryFile
from typing import Iterable, Literal

from handbrake.models.common import Offset
from handbrake.models.preset import Preset
from handbrake.models.progress import Progress
from handbrake.models.title import TitleSet
from handbrake.models.version import Version
from handbrake.progresshandler import ProgressHandler
from handbrake.runner import CommandRunner, OutputProcessor


class HandBrake:
    def __init__(self, executable: str = "HandBrakeCLI"):
        """Initialise the HandBrake wrapper

        :param executable: path to the HandBrakeCLI executable to use
        """
        self.executable = executable

    def version(self) -> Version:
        """Returns the version of HandBrakeCLI

        :returns: an object holding the handbrake version
        """
        version_processor = OutputProcessor(
            (b"Version: {", b"{"),
            (b"}", b"}"),
            Version.model_validate_json,
        )
        version: Version | None = None
        runner = CommandRunner(version_processor)
        cmd = [self.executable, "--json", "--version"]
        for obj in runner.process(cmd):
            if isinstance(obj, Version):
                version = obj

        if version is None:
            raise RuntimeError("version not found")
        return version

    def convert_title(
        self,
        input: str | PathLike,
        output: str | PathLike,
        title: int | Literal["main"],
        chapters: int | tuple[int, int] | None = None,
        angle: int | None = None,
        previews: tuple[int, bool] | None = None,
        start_at_preview: int | None = None,
        start_at: Offset | None = None,
        stop_at: Offset | None = None,
        audio: int | Iterable[int] | Literal["all", "first", "none"] | None = None,
        subtitles: (
            int | Iterable[int] | Literal["all", "first", "scan", "none"] | None
        ) = None,
        preset: str | None = None,
        preset_files: Iterable[str | PathLike] | None = None,
        presets: Iterable[Preset] | None = None,
        preset_from_gui: bool = False,
        no_dvdnav: bool = False,
        progress_handler: ProgressHandler | None = None,
    ):
        """Convert a title from the input source

        :param input: path to the input source
        :param output: path to the output file
        :param title: the index of the title to rip, or the literal string "main"
        to automatically select the main feature
        :param chapters: the chapter, or chapter range specified as (start, stop), to convert
        :param angle: the video angle to convert
        :param previews: an (int, bool) tuple indicating the number of previews to generate
        and whether the previews should be stored to disk
        :param start_at_preview: an index of the preview to start the conversion at
        :param start_at: an offset from the beginning of the media to start conversion at
        :param stop_at: an offset from the start_at parameter to stop conversion at
        :param audio: select which audio track(s) to convert
        :param subtitles: select which subtitle track(s) to convert
        :param preset: the name of the preset to use
        :param preset_files: a list of extra preset files to load when searching for the named preset
        :param presets: a list of `Preset` objects to load when searching for the named preset
        :param preset_from_gui: whether to import preset settings from the GUI
        :param no_dvdnav: switch to toggle whether to use dvdnav for reading DVDs
        :param progress_handler: a callback function to handle progress updates
        """
        if title == 0:
            raise ValueError("invalid title")

        with ExitStack() as stack:
            # generate list of preset import files
            preset_import_files: list[str] = []
            if preset_files is not None:
                preset_import_files += [str(f) for f in preset_files]
            if presets is not None:
                # generate a temporary file for each in-memory preset
                for p in presets:
                    f = NamedTemporaryFile("w")
                    stack.enter_context(f)
                    f.write(p.model_dump_json(by_alias=True))
                    f.flush()
                    preset_import_files.append(f.name)

            # generate command
            cmd: list[str] = [
                self.executable,
                "--json",
                "-i",
                str(input),
                "-o",
                str(output),
            ]
            if len(preset_import_files) > 0:
                cmd += [
                    "--preset-import-file",
                    " ".join(preset_import_files),
                ]
            if preset is not None:
                cmd += ["--preset", preset]
            if preset_from_gui:
                cmd += ["--preset-import-gui"]
            if title == "main":
                cmd += ["--main-feature"]
            else:
                cmd += ["-t", str(title)]
            if no_dvdnav:
                cmd += ["--no-dvdnav"]
            if isinstance(chapters, tuple):
                cmd += ["-c", f"{chapters[0]}-{chapters[1]}"]
            elif isinstance(chapters, int):
                cmd += ["-c", str(chapters)]
            if angle is not None:
                cmd += ["--angle", str(angle)]
            if previews is not None:
                cmd += ["--previews", f"{previews[0]}:{int(previews[1])}"]
            if start_at_preview is not None:
                cmd += ["--start-at-preview", str(start_at_preview)]
            if start_at is not None:
                cmd += ["--start-at", f"{start_at.unit}:{start_at.count}"]
            if stop_at is not None:
                cmd += ["--stop-at", f"{stop_at.unit}:{stop_at.count}"]
            if isinstance(audio, int):
                cmd += ["--audio", str(audio)]
            elif audio == "all":
                cmd += ["--all-audio"]
            elif audio == "first":
                cmd += ["--first-audio"]
            elif audio == "none":
                cmd += ["--audio", "none"]
            elif audio is not None:
                cmd += ["--audio", ",".join(str(a) for a in audio)]
            if isinstance(subtitles, int):
                cmd += ["--subtitle", str(subtitles)]
            elif subtitles == "all":
                cmd += ["--all-subtitles"]
            elif subtitles == "first":
                cmd += ["--first-subtitle"]
            elif subtitles == "none":
                cmd += ["--subtitle", "none"]
            elif subtitles == "scan":
                cmd += ["--subtitle", "scan"]
            elif subtitles is not None:
                cmd += ["--subtitle", ",".join(str(s) for s in subtitles)]

            # run command
            progress_processor = OutputProcessor(
                (b"Progress: {", b"{"),
                (b"}", b"}"),
                Progress.model_validate_json,
            )
            runner = CommandRunner(progress_processor)
            for obj in runner.process(cmd):
                if isinstance(obj, Progress):
                    if progress_handler is not None:
                        progress_handler(obj)

    def scan_title(
        self,
        input: str | PathLike,
        title: int | Literal["main"],
        progress_handler: ProgressHandler | None = None,
    ) -> TitleSet:
        """Scans the selected title and returns information about it

        :param input: path to the input source
        :param title: the index of the title to scan, or the literal string "main"
        to automatically select the main feature
        :param progress_handler: a callback function to handle progress updates
        :return: a `TitleSet` containing the selected title
        """
        if title == 0:
            raise ValueError("title == 0, use scan_all_titles to select all titles")

        # generate command
        cmd: list[str] = [self.executable, "--json", "-i", str(input), "--scan"]
        if title == "main":
            cmd += ["--main-feature"]
        else:
            cmd += ["-t", str(title)]

        # run command
        progress_processor = OutputProcessor(
            (b"Progress: {", b"{"),
            (b"}", b"}"),
            Progress.model_validate_json,
        )
        titleset_processor = OutputProcessor(
            (b"JSON Title Set: {", b"{"),
            (b"}", b"}"),
            TitleSet.model_validate_json,
        )
        title_set: TitleSet | None = None
        runner = CommandRunner(progress_processor, titleset_processor)
        for obj in runner.process(cmd):
            if isinstance(obj, Progress):
                if progress_handler is not None:
                    progress_handler(obj)
            elif isinstance(obj, TitleSet):
                title_set = obj

        # check output
        if title_set is None:
            raise RuntimeError("no titles found")
        if len(title_set.title_list) == 0:
            raise RuntimeError("title not found")
        return title_set

    def scan_all_titles(
        self,
        input: str | PathLike,
        progress_handler: ProgressHandler | None = None,
    ) -> TitleSet:
        """Scan all titles and return information about them

        :param input: path to the input source
        :param progress_handler: a callback function to handle progress updates
        :return: a `TitleSet` containing the all titles in the input source
        """
        # generate command
        cmd: list[str] = [
            self.executable,
            "--json",
            "-i",
            str(input),
            "--scan",
            "-t",
            "0",
        ]

        # run command
        progress_output_handler = OutputProcessor(
            (b"Progress: {", b"{"),
            (b"}", b"}"),
            Progress.model_validate_json,
        )
        titleset_output_handler = OutputProcessor(
            (b"JSON Title Set: {", b"{"),
            (b"}", b"}"),
            TitleSet.model_validate_json,
        )
        title_set: TitleSet | None = None
        runner = CommandRunner(progress_output_handler, titleset_output_handler)
        for obj in runner.process(cmd):
            if isinstance(obj, Progress):
                if progress_handler is not None:
                    progress_handler(obj)
            elif isinstance(obj, TitleSet):
                title_set = obj

        # check output
        if title_set is None:
            raise RuntimeError("no titles found")
        return title_set

    def get_preset(self, name: str) -> Preset:
        """Get the builtin preset with the given name

        :param name: the name of the preset to select
        :returns: a `Preset` object containing the selected preset
        """
        preset_list_processor = OutputProcessor(
            (b"{", b"{"), (b"}", b"}"), lambda d: Preset.model_validate_json(d)
        )
        preset_list: Preset | None = None
        runner = CommandRunner(preset_list_processor)
        cmd = [
            self.executable,
            "--json",
            "-Z",
            name,
            "--preset-export",
            name,
        ]
        for obj in runner.process(cmd):
            if isinstance(obj, Preset):
                preset_list = obj

        if preset_list is None:
            raise RuntimeError("no preset list found")
        return preset_list

    def list_presets(self) -> dict[str, dict[str, str]]:
        """List all builtin presets

        :returns: a dict[group, dict[name, description]] containing all builtin presets
        """
        res: dict[str, dict[str, str]] = {}
        group: dict[str, str] = {}
        preset: str = ""
        cmd = [self.executable, "-z"]
        proc = subprocess.run(cmd, capture_output=True, check=True)
        for line in proc.stderr.decode().splitlines():
            if line.endswith("/"):
                group = {}
                res[line[:-1]] = group
            elif line.startswith("        "):
                if group[preset] == "":
                    group[preset] = line.strip()
                else:
                    group[preset] += " " + line.strip()
            elif line.startswith("    "):
                preset = line.strip()
                group[preset] = ""
        return res

    def load_preset_from_file(self, file: str | PathLike) -> Preset:
        """Load a handbrake preset export into a `Preset` object

        :returns: a `Preset` object from the data in the given file
        """
        with open(file) as f:
            return Preset.model_validate_json(f.read())



import subprocess
from typing import Any, Callable, Generator, Generic, TypeVar

from handbrake.errors import HandBrakeError

T = TypeVar("T")


class OutputProcessor(Generic[T]):
    """
    Match the beginning and end of an object in command output and
    convert it to a model
    """

    def __init__(
        self,
        start_line: tuple[bytes, bytes],
        end_line: tuple[bytes, bytes],
        converter: Callable[[bytes], T],
    ):
        self.start_line = start_line
        self.end_line = end_line
        self.converter = converter

    def match_start(self, line: bytes) -> bytes | None:
        if line == self.start_line[0]:
            return self.start_line[1]
        return None

    def match_end(self, line: bytes) -> bytes | None:
        if line == self.end_line[0]:
            return self.end_line[1]
        return None

    def convert(self, data: bytes) -> T:
        return self.converter(data)


class CommandRunner:
    def __init__(self, *processors: OutputProcessor):
        self.processors = processors
        self.current_processor: OutputProcessor | None = None
        self.collect: list[bytes] = []

    def process_line(self, line: bytes) -> Any:
        if self.current_processor is None:
            # attempt to start a processor
            for processor in self.processors:
                c = processor.match_start(line)
                if c is not None:
                    self.current_processor = processor
                    self.collect = [c]
                    return
        else:
            # attempt to end the current processor
            c = self.current_processor.match_end(line)
            if c is not None:
                self.collect.append(c)
                res = self.current_processor.convert(b"\n".join(self.collect))
                self.current_processor = None
                self.collect = []
                return res
            # append line to current collect
            self.collect.append(line)

    def process(self, cmd: list[str]) -> Generator[Any, None, None]:
        # create process with pipes to output
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )
        if proc.stdout is None:
            raise ValueError

        # slurp stdout line-by-line
        while True:
            stdout = proc.stdout.readline().rstrip()
            if len(stdout) == 0 and proc.poll() is not None:
                break
            o = self.process_line(stdout)
            if o is not None:
                yield o

        # raise error on nonzero return code
        if proc.returncode != 0:
            raise HandBrakeError(proc.returncode)



from typing import Any

from handbrake.models.common import HandBrakeModel


class Preset(HandBrakeModel):
    version_major: int
    version_micro: int
    version_minor: int

    preset_list: list[dict[str, Any]]

    @property
    def version_string(self) -> str:
        return f"{self.version_major}.{self.version_minor}.{self.version_micro}"



from datetime import timedelta
from typing import Literal

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_pascal


class HandBrakeModel(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_pascal,
        populate_by_name=True,
    )


class Duration(HandBrakeModel):
    hours: int
    minutes: int
    seconds: int
    ticks: int

    def to_timedelta(self) -> timedelta:
        return timedelta(hours=self.hours, minutes=self.minutes, seconds=self.seconds)


class Fraction(HandBrakeModel):
    den: int
    num: int

    def to_float(self) -> float:
        return self.num / self.den


class Offset(HandBrakeModel):
    count: int
    unit: Literal["seconds"] | Literal["frames"] | Literal["pts"]
